using System;
using DerivativeFunctions;
using LinearAlgebra;

namespace Steppers
{
	/// <summary>
	/// A one-step method to solve an ordinary explicit differential equation of order 1.
	/// </summary>
	public interface Stepper
	{
		VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h);
	}

	public class Euler : Stepper
	{
		public VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h)
		{
			return v + h*f.Evaluate(t, v);
		}
	}

	public class Midpoint : Stepper
	{
		public VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h)
		{
			VectorN v1 = h*f.Evaluate(t        , v         );
			VectorN v2 = h*f.Evaluate(t + 0.5*h, v + 0.5*v1);
			return v + v2;
		}
	}

	public class Heun : Stepper
	{
		public VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h)
		{
			VectorN v1 = h*f.Evaluate(t              , v               );
			VectorN v2 = h*f.Evaluate(t + h*(1.0/3.0), v + v1*(1.0/3.0));
			VectorN v3 = h*f.Evaluate(t + h*(2.0/3.0), v + v2*(2.0/3.0));
			return v + 0.25*v1 + 0.75*v3;
		}
	}

	public class RungeKutta4 : Stepper
	{
		public VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h)
		{
			VectorN v1 = h*f.Evaluate(t        , v         );
			VectorN v2 = h*f.Evaluate(t + 0.5*h, v + 0.5*v1);
			VectorN v3 = h*f.Evaluate(t + 0.5*h, v + 0.5*v2);
			VectorN v4 = h*f.Evaluate(t +     h, v +     v3);
			return v + (v1 + 2.0*v2 + 2.0*v3 + v4)/6.0;
		}
	}

	public class SemiImplicitEuler : Stepper
	{
		public VectorN GetNext(DerivativeFunction f, double t, VectorN v, double h)
		{
			return v + h*(MatrixN.Identity(v.Size)-h*f.Gradient(t+h, v)).Invert()*f.Evaluate(t, v);
		}
	}
}
